#ifndef DV_TICKET_REQUEST_H
#define DV_TICKET_REQUEST_H

#include <dvxml/xml.h>
#include <dvssl/sslsocket.h>
#include <dvmysql/sqldb.h>
#include <dvticket/authenticator.h>

namespace Dv {
namespace Ticket {
class Server; 
/** A class representing a reques from a client. A client may e.g.
 * request to obtain a new ticket (anonymous or otherwise), or
 * to validate an existing ticket (via its serial number).
 * @sa Dv::Ticket::Ticket
 */
class Request {
public:
  /** Shorthand typedef */
  typedef Dv::Util::ref<Request> Ref;
  /** Factory function. The function will return a pointer to
   * an instance object of a class derived from a Request.
   * The derived class depends on the data in the req XML document
   * parameter.
   *
   * @param req XML document as received from a client.
   * @param client SSL connection with client making the request
   * @param server that received this request
   * @return 0 upon any error, e.g. unknown request type,
   *   otherwise a pointer to a valid instance of a class derived from Request.
   * @sa Dv::Ticket::LoginRequest, Dv::Ticket::AnonLoginRequest, Dv::Ticket::ValidateRequest
   */
  static Ref make(const Dv::Xml::Document& req, 
    Dv::Util::ref<Dv::Ssl::SslSocket> client, Server& server);
  /** Handle a request. This is a virtual function, the default
   * implementation simly returns its unmodified argument.
   * @param result reference to an XML document that will
   *   be filled with the reply to the request.
   * @return a reference to the first argument.
   */
  virtual Dv::Xml::Document& handle(Dv::Xml::Document& result) {
    return result;
    }
  /** Destructor */
  virtual ~Request() {}
  /** @return reference to root node of the request XML representation. */
  const Dv::Xml::Node::Ref& root() const { return root_; }
  /** @return dot address of client making the request, e.g. "134.184.65.2" */
  const std::string& host() const { return host_; }
  /** Server that received the request. */
  Dv::Ticket::Server& server() { return server_; }
protected:
  /** Constructor (protected, use Dv::Request::make).
   * @param req XML document as received from a client.
   * @param client SSL connection with client making the request
   * @param server that received this request
   * @sa Dv::Ticket::Request::make
   */
  Request(const Dv::Xml::Document& req, Dv::Util::ref<Dv::Ssl::SslSocket> client, Server& server);
  /** @return reference to SSL connection with requesting client */
  Dv::Util::ref<Dv::Ssl::SslSocket> client() { return client_; }
private:
  /** XML document containing the request */
  Dv::Xml::Document req_;
  /** root of req_ */
  Dv::Xml::Node::Ref root_;
  /** SSL connection to client making the request */
  Dv::Util::ref<Dv::Ssl::SslSocket> client_;
  /** Dot address of client making the request. E.g. 134.184.65.2 */
  std::string host_;
  /** Server that received the request. */
  Server& server_;
};

/** Request from a user to obtain a new ticket. */
class LoginRequest: public Request {
public:
  /** Constructor.
   * @param req XML document as received from a client.
   * @param client SSL connection with client making the request
   * @param server that received this request
   * @sa Dv::Ticket::Request::make
   */
  LoginRequest(const Dv::Xml::Document& req, Dv::Util::ref<Dv::Ssl::SslSocket> client,
    Server& server): Request(req, client, server) {}
  /** Handle a request. This is implements Dv::Ticket::Request::handle
   * @param result reference to an XML document that will
   *   be filled with the reply to the request.
   * @return a reference to the first argument.
   */
  Dv::Xml::Document& handle(Dv::Xml::Document& result);
};

/** Request from a user to obtain an anonymous ticket. */
class AnonLoginRequest: public Request {
public:
  /** Constructor.
   * @param req XML document as received from a client.
   * @param client SSL connection with client making the request
   * @param server that received this request
   * @sa Dv::Ticket::Request::make
   */
  AnonLoginRequest(const Dv::Xml::Document& req, Dv::Util::ref<Dv::Ssl::SslSocket> client,
    Server& server): Request(req, client, server) {}
  /** Handle a request. This is implements Dv::Ticket::Request::handle
   * @param result reference to an XML document that will
   *   be filled with the reply to the request.
   * @return a reference to the first argument.
   */
  Dv::Xml::Document& handle(Dv::Xml::Document& result);
};

/** Request from a user to validate a ticket. */
class ValidateRequest: public Request {
public:
  /** Constructor.
   * @param req XML document as received from a client.
   * @param client SSL connection with client making the request
   * @param server that received this request
   * @sa Dv::Ticket::Request::make
   */
  ValidateRequest(const Dv::Xml::Document& req, Dv::Util::ref<Dv::Ssl::SslSocket> client,
    Server& server): Request(req, client, server) {}
  /** Handle a request. This is implements Dv::Ticket::Request::handle
   * @param result reference to an XML document that will
   *   be filled with the reply to the request.
   * @return a reference to the first argument.
   */
  Dv::Xml::Document& handle(Dv::Xml::Document& result);
};

}}
#endif

