// $Id: server.C,v 1.14 2003/09/18 13:02:25 dvermeir Exp $

#include <iostream>
#include <stdexcept>
#include <dvxml/xml.h>
#include <dvmysql/mysqldb.h>
#include <dvutil/logstream.h>
#include <dvutil/ntostr.h>
#include <dvutil/random.h>
#include <dvutil/tostring.h>
#include <dvutil/file.h>
#include <dvutil/makedaemon.h>
#include <dvnet/inetaddress.h>
#include <dvssl/sslserversocket.h>

#include "xml-sndrcv.h"
#include "request.h"
#include "server.h"
#include "ticket.h"

Dv::Ticket::Server::~Server() {
delete log_;
delete db_;
delete ssl_ss_;
delete ssl_context_;
}

Dv::Ticket::Server::Server(const Dv::Xml::Node::Ref config, 
    Dv::Ticket::Authenticator& auth) throw (Dv::Ticket::Exception):
      authenticator_(auth), db_(0), log_(0),
      ssl_context_(0), ssl_ss_(0) {
init(config);
}

Dv::Ticket::Server::Server(const std::string& config_fn, 
    Dv::Ticket::Authenticator& auth) throw (Dv::Ticket::Exception):
      authenticator_(auth), db_(0), log_(0), ssl_context_(0), ssl_ss_(0) {
try {
  Dv::Util::File config_f(config_fn);
  if (! config_f.exists() )
    throw std::runtime_error(config_fn + ": cannot open");
  Dv::Xml::Document config(config_f);

  std::string config_dtd_fn = (config.root()/"dtd"/"config");
  
  Dv::Util::File dtd_f(config_dtd_fn);
  if (! dtd_f.exists() )
    throw std::runtime_error(config_dtd_fn + ": cannot open");
  config_dtd_fn = dtd_f.str();

  if (! config.valid(config_dtd_fn) )
    throw std::runtime_error(config_fn + ": not valid w.r.t. " + config_dtd_fn);

  init(config.root());

  log() << "  config file: " << config_fn << std::endl
        << "  config dtd file: " << config_dtd_fn << std::endl;
  }
catch (Dv::Ticket::Exception& e) {
  throw;
  }
catch (std::exception& e) {
  throw Dv::Ticket::Exception(e.what());
  }
}

void
Dv::Ticket::Server::init(const Dv::Xml::Node::Ref config) 
  throw (Dv::Ticket::Exception) {
try {
  Dv::Util::File request_dtd_f((config/"dtd"/"request").str());
  if (! request_dtd_f.exists() )
    throw std::runtime_error(request_dtd_f.str() + ": not found");
  request_dtd_fn_ = request_dtd_f.str();

  pid_fn_ = Dv::Util::File::absolute_path(config/"pidfile");

  // Open database
  std::string db_name = (config/"database"/"name");
  std::string db_user = (config/"database"/"user"/"name");
  std::string db_pass = (config/"database"/"user"/"passwd");
  
  db_ = new Dv::MySql::Db(db_name.c_str(), db_user.c_str(), db_pass.c_str() );
  
  if (!*db_)
    throw std::runtime_error(db_name + ": cannot open as " + db_user);
  
  // Open server socket.
  
  Dv::Util::File cert_f((config/"server"/"cert").str());
  if (! cert_f.exists() )
    throw std::runtime_error(cert_f.str() + ": not found");
  cert_fn_ = cert_f.str();

  Dv::Util::File key_f((config/"server"/"key").str());
  if (! key_f.exists() )
    throw std::runtime_error(key_f.str() + ": not found");
  key_fn_ = key_f.str();

  port_ = (config/"server"/"host"/"port");
  host_ = Dv::Net::InetAddress::local_host();
  if (host_ == "localhost")
    host_ = Dv::Net::InetAddress::by_name(host_)->dot_address();

  client_delay_ =  (config/"server"/"client-delay");
  
  ssl_context_ = new Dv::Ssl::SslContextV23(key_fn().c_str(), cert_fn().c_str());
  ssl_ss_ = new Dv::Ssl::SslServerSocket(*ssl_context_, port());

  // Open logfile.


  log_fn_ = Dv::Util::File::absolute_path(config/"logfile");
  log_ = new Dv::Util::logstream(log_fn(), "dvticketserver");

  if (!*log_)
    throw std::runtime_error(log_fn() + ": cannot open");

  log() << "ticketserver configured:" << std::endl
      << "  port = " << port() << std::endl
      << "  request dtd file = " << request_dtd_fn() << std::endl
      << "  client-delay = " << client_delay() << " millisecs" << std::endl
      << "  db: name = " << db_name << ", user = " << db_user << std::endl
      << "  certificate file = " << cert_fn() << std::endl
      << "  key file = " << key_fn() << std::endl
      << "  pid in " << pid_fn() << std::endl
      << "  log file: " << log_fn() << std::endl
      << "  authenticator: " << authenticator_.name() << std::endl;
  }
catch (std::exception& e) {
  throw Dv::Ticket::Exception(e.what());
  }
}

int
Dv::Ticket::Server::main() throw () {

// Become daemon.

try {
  log() << "pid_fn = " << pid_fn() << std::endl;
  Dv::Util::make_daemon(pid_fn());
  log() << "ticketserver started" << std::endl;

  // Server loop.

  while (true) {
    Dv::Util::ref<Dv::Ssl::SslSocket> client(ssl_ss_->accept(client_delay()));
    if (! *client) {
      log() << "accept() fails: " << client->strerror() << std::endl;
      continue;
      }
    log() << "connection from " << client->host() << "." << client->port() << std::endl;
    Dv::Xml::Document request_doc;
    Dv::Xml::receive(request_doc, *client);
    log() << "received:\n" << request_doc << std::endl;
    if (! request_doc.valid(request_dtd_fn()) ) {
      log() << "invalid request:\n" << request_doc << std::endl;
      continue;
      }

    try {
      Dv::Ticket::Request::Ref request(
        Dv::Ticket::Request::make(request_doc, client, *this)
	);
      if (!request) {
        log() << "cannot handle request:\n" << request_doc << std::endl;
	continue;
        }

      Dv::Xml::Document reply;
      Dv::Xml::send(request->handle(reply), *client);
      log() << "reply:\n" << reply << std::endl;
      }
    catch (std::exception& e) {
      log() << e.what() << std::endl;
      continue;
      }
    }
  }
catch (std::exception& e) {
  log() << e.what() << std::endl;
  return 1;
  }
return 0;
}

Dv::Ticket::Ticket::Ref
Dv::Ticket::Server::make_ticket(const Dv::Ticket::User& user, 
  const std::string& owner, size_t duration) throw (Dv::Ticket::Exception) {
try {
  Dv::Util::Date expiration_date;
  static const size_t MAX_TRIES(10);
  expiration_date.minutes(expiration_date.minutes() + duration);
  
  // MYSQL date format %Y-%m-%d  %H:%M:%S
  std::string  expires(expiration_date.str("%Y-%m-%d %H:%M:%S"));
  
  std::string dotowner(Dv::Net::InetAddress::by_name(owner)->dot_address());

  bool inserted(false);
  std::string error_msg;
  
  Dv::Ticket::Ticket::SERIAL serial(0);
  for (size_t i=0; ( (i<MAX_TRIES) && !inserted); ++i) {
    serial = Dv::Util::Random::rand();
  
    Dv::Sql::Command u(db(), "insert into ticket (id,host,expiration,uid,utype,name,info) values");
    u << "(" << serial << "," << dotowner << "," << expires << "," 
      << user.id() << "," << user.category() << "," << user.name() 
      << "," << user.info() << ")";
  
    inserted =  u.exec();
    error_msg = u.err();
    }
  
  if (! inserted)
    throw std::runtime_error(error_msg);

  return Dv::Ticket::Ticket::Ref(
    new Dv::Ticket::Ticket(user, serial, expiration_date, host(), port(), dotowner)
    );
  }
catch (std::exception& e) {
  static const std::string FUNC("while storing new ticket: ");
  throw Dv::Ticket::Exception(FUNC + e.what());
  }
}

Dv::Ticket::Ticket::Ref
Dv::Ticket::Server::validate_ticket(Dv::Ticket::Ticket::SERIAL serial, 
  const std::string& owner) throw (Dv::Ticket::Exception) {
try {
  std::string dotowner(Dv::Net::InetAddress::by_name(owner)->dot_address());
  Dv::Sql::Command  q(db(), "select expiration,uid,utype,name,info from ticket where id = ");
  q << serial << "and host=" << dotowner;
  
  if (! q.exec() ) 
    throw std::runtime_error(q.err());
  
  if ( q.begin()==q.end() ) 
    throw std::runtime_error(
      Dv::Util::tostring(serial) + ": unknown ticket, host = " + dotowner
      );
  
  Dv::Sql::Command::iterator it(q.begin());
  Dv::Util::Date now;
  Dv::Util::Date expiration_date(it(0));
  
  if (now >= expiration_date) 
    throw std::runtime_error(
      Dv::Util::tostring(serial) + " expired on " + expiration_date.str("%Y-%m-%d %H:%M:%S") 
      );
  
  Dv::Ticket::User user(it(2), Dv::Util::Int(it(1)), it(3), it(4));
  return Dv::Ticket::Ticket::Ref(new Dv::Ticket::Ticket(user, serial, expiration_date, host(), port(), dotowner));
  }
catch (std::exception& e) {
  static const std::string FUNC("while retrieving ticket: ");
  throw Dv::Ticket::Exception(FUNC + e.what());
  }
}
