#ifndef	DVTICKET_SERVER_H
#define	DVTICKET_SERVER_H

// $Id: server.h,v 1.7 2003/09/18 13:02:25 dvermeir Exp $

#include <string>
#include <iostream>
#include <stdexcept>
#include <dvxml/xml.h>
#include <dvmysql/mysqldb.h>
#include <dvssl/sslserversocket.h>
#include <dvticket/authenticator.h>
#include <dvticket/ticket.h>
#include <dvticket/exception.h>

namespace Dv { 
namespace Ticket {

/** Ticket server class. 
 * @warning The server is single-threaded. DOS attacks are --
 * hopefully -- avoided by using a max. client delay.
 * @sa Dv::Ticket::Server::client_delay
 */
class Server {
public:
  /** Constructor.
   * @param config XML node containing configuration parameters. This
   *   node is assumed to conform to the configuration DTD.
   * @param authenticator object to which actual user name/passwd
   *   verification will be delegated.
   * @exception Dv::Ticket::Exception upon any error
   *
   * The constructor will connect to the database, set up
   * the SSL server etc.
   */
  Server(const Dv::Xml::Node::Ref config, Dv::Ticket::Authenticator& authenticator)
    throw (Dv::Ticket::Exception);

  /** Constructor.
   * @param config_fn name of XML file containing configuration
   *   parameters. The file will be checked vs the DTD.
   * @param authenticator object to which actual user name/passwd
   *   verification will be delegated.
   * @exception Dv::Ticket::Exception upon any error
   *
   * The constructor will connect to the database, set up
   * the SSL server etc.
   */
  Server(const std::string& config_fn, Dv::Ticket::Authenticator& authenticator)
    throw (Dv::Ticket::Exception);

  /** Server main loop.
   * @return 0 if ok, non-zero if error.
   */
  int main() throw ();

  /** Destructor (virtual). */
  virtual ~Server();

  /** @return name of directory containin DTD files */
  const std::string& dtd_dir() const { return dtd_dir_; }
  /** @return name of configuration file */
  const std::string& config_fn() const { return config_fn_; }
  /** @return name of configuration DTD file */
  const std::string& config_dtd_fn() const { return config_dtd_fn_; }
  /** @return name of request DTD file */
  const std::string& request_dtd_fn() const { return request_dtd_fn_; }
  /** @return name of log file */
  const std::string& log_fn() const { return log_fn_; }
  /** @return name of server certificate file */
  const std::string& cert_fn() const { return cert_fn_; }
  /** @return name of server key file */
  const std::string& key_fn() const { return key_fn_; }
  /** @return name of file containing process ID of this server */
  const std::string& pid_fn() const { return pid_fn_; }

  /** @return port number on which the server listens for connections */
  int port() const { return port_; }
  /** @return host name, in dot notation, of the server (this machine). */
  const std::string& host() const { return host_; }

  /** @return delay the server is prepared to wait for any
   * communication from the client, in millisec */
  size_t client_delay() const { return client_delay_; }

  /** @return reference to log output stream */
  std::ostream& log() { return *log_; }
  /** @return pointer to ticket database connection, never 0 */
  Dv::MySql::Db& db() { return *db_; }

  /** @return reference to authentictor object that is used by this server */
  Dv::Ticket::Authenticator& authenticator() { return authenticator_; }

  /** Create and store a new ticket for a user.
   * @param user which is supposed to have been authenticated
   * @param host for which ticket will be valid, in dot notation (e.g.  134.184.65.2)
   * @param duration that ticket will be valid, in minutes.
   * @exception Dv::Ticket::Exception if anything goes wrong
   */
  Dv::Ticket::Ticket::Ref make_ticket(const User& user, const std::string& host,
    size_t duration) throw (Dv::Ticket::Exception);

  /** Retrieve a ticket with a given serial number.
   * @param serial unique numeric ID of ticket
   * @param host that should be associated with the ticket
   * @exception Dv::Ticket::Exception if anything goes wrong,
   *  e.g. the ticket is no longer valid, or the ticket with the
   *  given id was issued for a different host.
   */
  Dv::Ticket::Ticket::Ref validate_ticket(Dv::Ticket::Ticket::SERIAL serial, const std::string& host)
    throw (Dv::Ticket::Exception);
  
private:

  /** Initialize a server.
   * @param config XML root node containing configuration info
   * @exception Dv::Ticket::Exception if anything goes wrong,
   *  e.g. config is not valid w.r.t config_dtd_fn.
   */
  void init(const Dv::Xml::Node::Ref config) throw (Dv::Ticket::Exception);

  /** name of directory containing relevant DTD files */
  std::string dtd_dir_;
  /** configuration file name */
  std::string config_fn_;
  /** configuration DTD file name */
  std::string config_dtd_fn_;
  /** request DTD file name */
  std::string request_dtd_fn_;
  /** log DTD file name */
  std::string log_fn_;
  /** server certificate file name */
  std::string cert_fn_;
  /** server key file name */
  std::string key_fn_;
  /** name of file containing server process ID */
  std::string pid_fn_;

  /** port on which server is listening */
  int port_;
  /** this host */
  std::string host_;
  /** delay that the server is prepared to wait for any
   * communication from the client, in millisec */
  size_t client_delay_;

  /** reference to authentictor object that is used by this server */
  Dv::Ticket::Authenticator& authenticator_;
  /** pointer to ticket database connection, never 0 */
  Dv::MySql::Db* db_;
  /** pointer to log output stream, never 0 */
  std::ostream* log_;
  /** SSL context */
  Dv::Ssl::SslContextV23* ssl_context_;
  /** SSL server socket */
  Dv::Ssl::SslServerSocket* ssl_ss_;
};

} }
#endif

