#ifndef	DV_TICKET_TICKET_H
#define	DV_TICKET_TICKET_H

// $Id: ticket.h,v 1.13 2003/09/18 13:02:25 dvermeir Exp $
#include <string>
#include <set>
#include <iostream>

#include <dvticket/user.h>
#include <dvutil/date.h>
#include <dvutil/ref.h>

namespace Dv {
namespace Ticket {

class Server;

/** Class representing a ticket. */
class Ticket {
friend class Server;

public:

  /** Type of serial number of ticket. */
  typedef unsigned long SERIAL;
  /** Type of reference-counted pointer to ticket. */
  typedef Dv::Util::ref<Dv::Ticket::Ticket> Ref;

  /** @return Owner of this ticket. */
  const User&  user() const { return user_; }
  /** @return Serial number of this ticket. */
  SERIAL serial() const { return serial_; }
  /** @return expiration date of this ticket */
  const Dv::Util::Date&  expiration_date() const { return expiration_date_; }
  /** @return host of server that issued this ticket */
  const std::string& server_host() const { return server_host_; }
  /** @return port of server that issued this ticket */
  const int server_port() const { return server_port_; }
  /** @return host for which this ticket is valid */
  const std::string& owner_host() const { return owner_host_; }

  /** Return an XML representation of a ticket.
   * @param name to be used as the root tag
   * @return root node of XML representation
   */
  Dv::Xml::Node xml(const std::string& name="ticket") const;

  /** Obtain a new ticket from a ticket server.
   * @param host name of ticket server
   * @param port on which ticket server is listening
   * @param name of user to obtain ticket for
   * @param password of user to obtain ticket for
   * @param info to be added to (user date member) in ticket
   * @param duration (in minutes) that ticket should remain valid.
   * @return a reference-counted pointer to a valid ticket,
   *   or a nil pointer.
   * @sa Dv::Ticket::Authenticator
   */
  static Ref buy(const std::string& host, int port, const std::string& name, 
    const std::string& password, const std::string& info="", size_t duration=30)
      throw ();

  /** Obtain a new ticket from a ticket server. This function
   * explicitly specifies the client host, which may be different
   * from the host on which this function is called.
   * It is useful e.g. for general "login" cgi programs. Such
   * programs avoid user programs to deal with passwords.
   * @param client_host name of ticket server
   * @param server_host name of ticket server
   * @param port on which ticket server is listening
   * @param name of user to obtain ticket for
   * @param password of user to obtain ticket for
   * @param info to be added to (user date member) in ticket
   * @param duration (in minutes) that ticket should remain valid.
   * @return a reference-counted pointer to a valid ticket,
   *   or a nil pointer.
   * @warning This function must be called from the same host
   *   as the one running the ticket server.
   * @sa Dv::Ticket::Authenticator
   */
  static Ref buy(const std::string& client_host, const std::string& server_host, int port,
    const std::string& name, const std::string& password, const std::string& info="", 
    size_t duration=30) throw ();

  /** Buy an anonymous ticket from a ticket server.
   * @param host name of ticket server
   * @param port on which ticket server is listening
   * @param name of user to obtain ticket for
   * @param info to be added to (user date member) in ticket
   * @param duration (in minutes) that ticket should remain valid.
   * @return a reference-counted pointer to a valid ticket,
   *   or a nil pointer. The ticket user will have "anonymous"
   *   as its category.
   * @sa Dv::Ticket::User
   * @sa Dv::Ticket::Authenticator
   */
  static Ref buy_anon(const std::string& host, int port, const std::string& name="anonymous", 
    const std::string& info="", size_t duration=30) throw ();


  /** Obtain a ticket by validating a serial number with a ticket
   * server.
   * @param server_host name of ticket server
   * @param server_port on which ticket server is listening
   * @param serial number of previously obtained ticket
   * @return a reference-counted pointer to a valid ticket,
   *   or a nil pointer.
   */
  static Ref validate(const std::string& server_host, int server_port, SERIAL serial)
    throw ();

  /** Obtain a ticket by validating a serial number with a ticket
   * server.
   * @param server_host name of ticket server
   * @param server_port on which ticket server is listening
   * @param host dot address of owner host of ticket
   * @param serial number of previously obtained ticket
   * @return a reference-counted pointer to a valid ticket,
   *   or a nil pointer.
   */
  static Ref validate(const std::string& server_host, int server_port,
    const std::string& host, SERIAL serial) throw ();

private:
  /** Copy ctor is forbidden. */
  Ticket(const Ticket&); 
  /** Assignment is forbidden. */
  Ticket& operator=(const Ticket&); 

  /** Simple ctor. 
   * @param user which is supposed to have been authenticated
   * @param serial unique numeric ID of ticket
   * @param expiration_date when ticket expires
   * @param server_host host of server issuing ticket
   * @param server_port host of server issuing ticket
   * @param owner_host host for which this ticket is valid (in dot notation)
   */
  Ticket(const User& user, SERIAL serial, Dv::Util::Date expiration_date,
    const std::string& server_host, int server_port, const std::string& owner_host) throw ():
      user_(user), serial_(serial), expiration_date_(expiration_date),
      server_host_(server_host), server_port_(server_port), owner_host_(owner_host) {}

  /** Simple ctor. 
   * @param root Xml node containing XML representation of ticket
   * @param server_host host of server issuing ticket
   * @param server_port host of server issuing ticket
   * @exception std::exception if xml node is not correct format
   */
  Ticket(const Dv::Xml::Node::Ref& root, const std::string& server_host, int server_port);


  /** User to whom this ticket was issued. */
  User  user_;
  /** Host for which this ticket was issued. */
  // std::string  host_;
  /** Unique serial number of this ticket. */
  SERIAL  serial_;
  /** Expiration date for this ticket. */
  Dv::Util::Date  expiration_date_;
  /** Host of server that issued this ticket. */
  std::string server_host_;
  /** Port of server that issued this ticket. */
  int server_port_;
  /** Name of host for which this ticket is valid. */
  std::string  owner_host_;
};

}}

#endif
